#!/usr/bin/perl -w
# class Data::Array2ArrayMap::SDB::Writer;
# Igor Vlasenko vlasenko@imath.kiev.ua

package Data::Array2ArrayMap::SDB::Writer;

use strict;
use Carp;

use constant VERSION=>123; # 1.2.3
use constant MAGIC_STRING=>'NSDB'; # magic value

use constant BYTES_IN_OFFSET_T=>4;
use constant SDB_HEADER_LENGTH=>6;

my $debug=0;

my $NRIF_HEADER_LENGTH=8;
my $NRIF_HEADER_OFFSET=BYTES_IN_OFFSET_T*$NRIF_HEADER_LENGTH;

use constant SDB_INT4	=>	2;
use constant SDB_INT8	=>	3;
use constant SDB_STRING4=>	4;
use constant SDB_CONTEXT4=>	6;
use constant SDB_DOUBLE	=>	8;

my %HAS_EXTERNAL_VALUES= map {$_=>1} 
    Data::Array2ArrayMap::SDB::Writer::SDB_INT8,
    Data::Array2ArrayMap::SDB::Writer::SDB_STRING4,
    Data::Array2ArrayMap::SDB::Writer::SDB_DOUBLE;

my $INFO_OFFSET=SDB_HEADER_LENGTH*BYTES_IN_OFFSET_T;
my $ROOT_NRIF_OFFSET=256+$INFO_OFFSET; # +0x100h

sub packoffset {return pack ("l*",@_)};

sub new {
    my $class=shift;
    my $filename=shift;
    my $self={
	DEBUG => $debug,
	FILENAME =>$filename,
	END_OF_FILE_OFFSET=>$ROOT_NRIF_OFFSET,
	ROOT_NRIF_OFFSET=>$ROOT_NRIF_OFFSET,
	NRIF_OFFSETS_HASH=>{},
    };
    return bless $self, $class;
}

sub create {
    my ($self)=@_;
    my $filename=$self->{FILENAME};
    # creating header
    open(IDXFILE,">$filename") || die "filesystem error $!";
    print STDERR "creating empty database $filename\n" if $self->{DEBUG};
    $self->write_info_string(\*IDXFILE);
    $self->write_root_nrif(\*IDXFILE);
    $self->write_header(\*IDXFILE);
    close (IDXFILE);
}

# private:
sub write_header {
    my ($self,$IDXFILE)=@_;

    # header of SDB --- SDB_HEADER_LENGTH=>6 values
    seek ($IDXFILE, 0, 0);
    ######################## 4 bytes #############################
    print $IDXFILE MAGIC_STRING; # magic value 4 bytes		#0
    # ------------------------------------------------------------
    print $IDXFILE pack("L",VERSION); # version 4 bytes  	#1
    # ------------------------------------------------------------
    print $IDXFILE packoffset($INFO_OFFSET);	#4 bytes	#2
    # ------------------------------------------------------------
    print $IDXFILE packoffset($self->{ROOT_NRIF_OFFSET});	#3
    # ------------------------------------------------------------
    # end of file offset (we write new nodes there)  #4 bytes   #4
    # usually it is the end of file, but if we delete last attribute
    # without truncating the file, this offset may lay inside the file
    print $IDXFILE packoffset($self->{END_OF_FILE_OFFSET});	#5
    # ------------------------------------------------------------

    print STDERR "END_OF_FILE=".$self->{END_OF_FILE_OFFSET}."\n" if $self->{DEBUG};
}

# private:
sub write_root_nrif {
    my ($self,$IDXFILE)=@_;
    my $offset=$self->{END_OF_FILE_OFFSET};
#    my $offset=$ROOT_NRIF_OFFSET;
    print STDERR "writing root NRIF at $offset\n" if $self->{DEBUG};
    my $nrif=Data::Array2ArrayMap::SDB::Writer::NRIF->new(
	[Data::Array2ArrayMap::SDB::Writer::SDB_INT4,Data::Array2ArrayMap::SDB::Writer::SDB_CONTEXT4],
	$self->{END_OF_FILE_OFFSET});
    my ($nrif_begin, $nrif_end)=$nrif->write_node(\*IDXFILE,$self->{NRIF_OFFSETS_HASH});
    $self->{ROOT_NRIF_OFFSET}=$nrif_begin;
    $self->{END_OF_FILE_OFFSET}=$nrif_end;
}

# private:
sub write_info_string {
    my ($self,$IDXFILE)=@_;
    # header 
    seek ($IDXFILE, $INFO_OFFSET, 0);
    my $filename=$self->{FILENAME};
    print $IDXFILE "NRIF SDB VER 1.2.3\n".
	"Created ".`date`."\n".
	"Name: $filename\n".
	"(C)2002-2007 Megakvartal\n\0".
	"(C)2003 Igor Vlasenko & Alexander Strelets\n\0";
	"(C)2007 Igor Vlasenko\n\0";
}

sub add_node {
    my $self=shift;
    my $key=shift;
    my $typeptr=shift;
    my $node=shift;
    my @FLAGS=@_;

    die "attribute $key is already used!!!" if $self->{NRIF_OFFSETS_HASH}->{$key};
    
    my $filename=$self->{FILENAME};
    open(IDXFILE,"+<$filename") || die "filesystem error $!";
    print STDERR "appending attribute $key to $filename\n" if $self->{DEBUG};

    #my $startpos=$self->{END_OF_FILE_OFFSET};
    my $startpos=$self->{ROOT_NRIF_OFFSET};
    my $nrif=Data::Array2ArrayMap::SDB::Writer::NRIF->new($typeptr,$startpos,@FLAGS);

    my ($nrif_begin,$nrif_end)=$nrif->write_node(\*IDXFILE,$node);

    $self->{END_OF_FILE_OFFSET}=$nrif_end;
    $self->{NRIF_OFFSETS_HASH}->{$key}=[$nrif_begin];

    $self->write_root_nrif(\*IDXFILE);
    $self->write_header(\*IDXFILE);

    close (IDXFILE);
}



#######################################################
#######################################################
#	SEPARATE PACKAGE Data::Array2ArrayMap::SDB::Writer::TableItem;	#
#-----------------------------------------------------#
#	describes an item to write into NRIF	      #
#######################################################
#######################################################

########################################
# note: offsets are relative to nrif   #
########################################


package Data::Array2ArrayMap::SDB::Writer::TableItem;

my $ItemKEY=0;
my $ItemVALUE=1;
my $ItemKEYENTRY=2;
my $ItemVALENTRY=3;

sub new {
    my $class=shift;
    my ($key,$value)=@_;
    return bless [$key,$value,$key,$value];
}

sub sort_by_value {
    my $type=shift;
    if ($type == Data::Array2ArrayMap::SDB::Writer::SDB_STRING4) {
	return sort { $a->[$ItemVALUE] cmp $b->[$ItemVALUE]} @_;
    } else {
	return sort { $a->[$ItemVALUE] <=> $b->[$ItemVALUE]} @_;
    }
}

sub sort_by_key {
    my $type=shift;
    return sort { $a->[$ItemKEY] <=> $b->[$ItemKEY]} @_;
}


sub add_key_offset {
    my $keytype=shift;
    my $itemsptr=shift;
    my $offset=shift;
    &__add_offset($keytype, $ItemKEYENTRY, $itemsptr, $offset);
}

sub add_val_offset {
    my $valtype=shift;
    my $itemsptr=shift;
    my $offset=shift;
    &__add_offset($valtype, $ItemVALENTRY, $itemsptr, $offset);
}

sub __add_offset {
    my $type=shift;
    my $field=shift;
    my $itemsptr=shift;
    my $offset=shift;
    if ($type == Data::Array2ArrayMap::SDB::Writer::SDB_STRING4
	or $type == Data::Array2ArrayMap::SDB::Writer::SDB_DOUBLE
	or $type == Data::Array2ArrayMap::SDB::Writer::SDB_INT8
	or $type == Data::Array2ArrayMap::SDB::Writer::SDB_CONTEXT4
	) {
	for (my $i=0; $i<@$itemsptr; $i++) {
	    my $item=$itemsptr->[$i];
	    $item->[$field]+=$offset;
	}
    } elsif ($type == Data::Array2ArrayMap::SDB::Writer::SDB_INT4
	     ) {
        # do nothing (raw value is stored);
    } else {
	die "unknown type $type!!";
    }
}

sub count_data_offsets {
    my ($keytype, $valtype, $itemsptr)=@_;
    my $keyoffset=0;
    my $valoffset=0;
    map {
	&data_offset($_,$keytype, $valtype, \$keyoffset,\$valoffset);
    } @$itemsptr;
    return ($keyoffset, $valoffset);
}

sub data_offset {
    my ($self,$keytype, $valtype, $keyoffsetptr,$valoffsetptr)=@_;
    &__data_offset($self, $ItemKEY, $ItemKEYENTRY, $keytype, $keyoffsetptr);
    &__data_offset($self, $ItemVALUE, $ItemVALENTRY, $valtype, $valoffsetptr);
}

#private
sub __data_offset {
    my ($self,$fielddata, $fieldentry, $type,$offsetptr)=@_;
    my $val=$self->[$fielddata];
    if ($type==Data::Array2ArrayMap::SDB::Writer::SDB_STRING4) {
	$self->[$fieldentry]=$$offsetptr;
	$$offsetptr+=length($val)+1; # \0 included
    } elsif ($type==Data::Array2ArrayMap::SDB::Writer::SDB_INT8) {
	$self->[$fieldentry]=$$offsetptr;
	$$offsetptr+=8; # LEN of int 8
    } elsif ($type==Data::Array2ArrayMap::SDB::Writer::SDB_DOUBLE) {
	$self->[$fieldentry]=$$offsetptr;
	$$offsetptr+=8; # LEN of double
    } elsif ($type==Data::Array2ArrayMap::SDB::Writer::SDB_CONTEXT4) {
	# raw value
	$self->[$fieldentry]=$val;
    } elsif ($type==Data::Array2ArrayMap::SDB::Writer::SDB_INT4) {
	# raw value
	$self->[$fieldentry]=$val;
    } else {
	die "unknown type $type!!";
    }
}

sub convert2val_entry {
    my $itemsptr=shift;
    my @entries=map {$_->[$ItemVALENTRY]} @$itemsptr;
    return \@entries;
}

sub convert2key_entry {
    my $itemsptr=shift;
    my @entries=map {$_->[$ItemKEYENTRY]} @$itemsptr;
    return \@entries;
}

sub convert2key {
    my $itemsptr=shift;
    my @keys=map {$_->[$ItemKEY]} @$itemsptr;
    return \@keys;
}

sub convert2val {
    my $itemsptr=shift;
    my @val=map {$_->[$ItemVALUE]} @$itemsptr;
    return \@val;
}




#######################################################
#######################################################
#	SEPARATE PACKAGE Data::Array2ArrayMap::SDB::Writer::NRIF		#
#-----------------------------------------------------#
#	describes an attribute inside StaticDB	      #
#######################################################
#######################################################


package Data::Array2ArrayMap::SDB::Writer::NRIF;
use Carp;

use vars qw/@IS_A/;
BEGIN {@IS_A='Data::Array2ArrayMap::SDB::Writer'};

use constant NRIF_MAGIC_STRING=>'NRIF'; # magic value
use constant BYTES_IN_OFFSET_T=>Data::Array2ArrayMap::SDB::Writer::BYTES_IN_OFFSET_T;

my %packsub = (
    Data::Array2ArrayMap::SDB::Writer::SDB_INT4 => sub {return pack ("l*",@_)},
    Data::Array2ArrayMap::SDB::Writer::SDB_CONTEXT4 => sub {return pack ("l*",@_)},
    Data::Array2ArrayMap::SDB::Writer::SDB_STRING4 => sub {return pack ("L*",@_)},
    # BUG!
    Data::Array2ArrayMap::SDB::Writer::SDB_INT8 => sub {return pack ("l*",@_)},
    # BUG!
    Data::Array2ArrayMap::SDB::Writer::SDB_DOUBLE => sub {return pack ("L*",@_)},
    );

# public
sub new {
    my $class=shift;
    my $typeptr=shift;
    my $header_offset=shift;

    confess "bad typepath supplied!" if not ref $typeptr or not $typeptr->[0] or not $typeptr->[1];

    my $self={
	KEYTYPE => $typeptr->[0],
	VALTYPE => $typeptr->[1],
	TYPEPATH => $typeptr,
	HEADER_OFFSET => $header_offset,
	END_OFFSET => undef, # not known yet
	HAS_FIND	=> 1,
	HAS_GET		=> 1,
	IS_GET_UNIQUE	=> 0,
	IS_FIND_UNIQUE	=> 0,
	DEFAULT_VALUE => undef,
	ELNUM		=> 0,
	DEBUG => $debug,
	@_};
#########################################
#        this keys are relative (commented because not initialized yet)
#	FIND_TABLE_OFFSET=> 0;
#	GET_TABLE_OFFSET=> 0;
#	VALUES_OFFSET	=> 0;
#	KEYS_OFFSET	=> 0;
#	END_OFFSET	=> 0;
#########################################
    return bless $self;
}

sub write_node {
    my $self =shift;
    my $IDXFILE=shift;
    my $node=shift;
    my @getorder;

    # todo: tests! raw hacking
    my @keys=sort {$a <=> $b} keys %$node;

    if (&is_leaf_node($node)) { 
	#print STDERR "leaf node dump:\n" if $self->{DEBUG};
	#use Data::Dumper;print STDERR Data::Dumper->new([$node])->Terse(1)->Indent(1)->Deepcopy(1)->Dump;
	foreach my $key (@keys) {
	    my @values=@{$node->{$key}};
	    foreach my $val (@values) {
		push @getorder, Data::Array2ArrayMap::SDB::Writer::TableItem->new($key,$val);
	    }
	}
    } else { # context node
	$self->{VALTYPE}=Data::Array2ArrayMap::SDB::Writer::SDB_CONTEXT4;
	my @subpath=@{$self->{TYPEPATH}};
	shift @subpath;
	foreach my $key (@keys) {
	    my $subnode=$node->{$key};
	    #print STDERR " start writing context $key at $self->{HEADER_OFFSET}\n" if $self->{DEBUG};
	    my @flags =(HAS_FIND	=> $self->{HAS_FIND},
			HAS_GET		=> $self->{HAS_GET},
			IS_GET_UNIQUE	=> $self->{IS_GET_UNIQUE},
			IS_FIND_UNIQUE	=> $self->{IS_FIND_UNIQUE},
			DEFAULT_VALUE => $self->{DEFAULT_VALUE},);

	    my $context_nrif=Data::Array2ArrayMap::SDB::Writer::NRIF->new(\@subpath, $self->{HEADER_OFFSET},
									  @flags);
	    my ($nrif_begin,$nrif_end)=$context_nrif->write_node($IDXFILE,$subnode);
	    push @getorder, Data::Array2ArrayMap::SDB::Writer::TableItem->new($key, $nrif_begin);
	    print STDERR "  created context entry for $key at $self->{HEADER_OFFSET}\n" if $self->{DEBUG};
	    #print STDERR "  new end_of_file $nrif_end\n" if $self->{DEBUG};
	    $self->{HEADER_OFFSET}=$nrif_end;
	}
    }

    if ($self->{VALTYPE} == Data::Array2ArrayMap::SDB::Writer::SDB_CONTEXT4) {
	$self->{HAS_FIND}=0;
	$self->{IS_GET_UNIQUE}=1;
	$self->{DEFAULT_VALUE}=undef;
    }

    $self->{ELNUM}=scalar @getorder;

    $self->calc_tables_offsets();

    my ($keysize, $valsize) = 
	&Data::Array2ArrayMap::SDB::Writer::TableItem::count_data_offsets($self->{KEYTYPE},$self->{VALTYPE},\@getorder);
    $self->{VALUES_OFFSET}+=$keysize;
    $self->{END_OFFSET}+=$keysize+$valsize;

    print STDERR "# got keysize, valsize = $keysize, $valsize\n" if $self->{DEBUG};


    my $add_keys_offset = $self->{KEYS_OFFSET};
    &Data::Array2ArrayMap::SDB::Writer::TableItem::add_key_offset($self->{KEYTYPE},\@getorder,$add_keys_offset);


    my $add_vals_offset = $self->{VALUES_OFFSET};
    $add_vals_offset= -$self->{HEADER_OFFSET} if $self->{VALTYPE} == Data::Array2ArrayMap::SDB::Writer::SDB_CONTEXT4;
    &Data::Array2ArrayMap::SDB::Writer::TableItem::add_val_offset($self->{VALTYPE},\@getorder,$add_vals_offset);
    
    print STDERR "start writing NRIF ($self->{ELNUM} elements) at $self->{HEADER_OFFSET} keytype=$self->{KEYTYPE} valtype=$self->{VALTYPE}\n" if $self->{DEBUG};

    $self->write_header($IDXFILE);
    $self->write_data($IDXFILE,\@getorder);

print STDERR "raw key entries to write:".join(' ',@{&Data::Array2ArrayMap::SDB::Writer::TableItem::convert2key_entry(\@getorder)}),"\n" if $self->{DEBUG};
print STDERR "raw val entries to write:".join(' ',@{&Data::Array2ArrayMap::SDB::Writer::TableItem::convert2val_entry(\@getorder)}),"\n" if $self->{DEBUG};

    if ($self->{HAS_GET}) {
	$self->write_table($IDXFILE, $self->{GET_TABLE_OFFSET},$self->{KEYTYPE},$self->{VALTYPE},
			   &Data::Array2ArrayMap::SDB::Writer::TableItem::convert2key_entry(\@getorder),
			   &Data::Array2ArrayMap::SDB::Writer::TableItem::convert2val_entry(\@getorder));
    }

    if ($self->{HAS_FIND}) {
	my @findorder=&Data::Array2ArrayMap::SDB::Writer::TableItem::sort_by_value($self->{VALTYPE},@getorder);
	$self->write_table($IDXFILE, $self->{FIND_TABLE_OFFSET},$self->{VALTYPE},$self->{KEYTYPE},
			   &Data::Array2ArrayMap::SDB::Writer::TableItem::convert2val_entry(\@findorder),
			   &Data::Array2ArrayMap::SDB::Writer::TableItem::convert2key_entry(\@findorder));
    }

    return $self->{HEADER_OFFSET}, $self->{HEADER_OFFSET}+$self->{END_OFFSET};
}

# private 
sub is_leaf_node {
    my $node=shift;
    my @values=values (%$node);
    # empty node always have no context;
    return 1 if 0==@values; 
    my $type=ref $values[0];

    return 1 if $type eq 'ARRAY';
    return 0 if $type eq 'HASH';
    return 0 if $type eq 'Data::Array2ArrayMap::Hash::Tree';
    die "unknown node type $type: $node->($values[0])!";
}

#private 
sub write_header {
    my ($self,$IDXFILE)=@_;

    my $HAS_DEFAULT_VALUE=defined $self->{DEFAULT_VALUE};
    my $DEFAULT_VALUE=0;
    if ($HAS_DEFAULT_VALUE) {
	my $defval=Data::Array2ArrayMap::SDB::Writer::TableItem->new(0, $self->{DEFAULT_VALUE});
	my ($keysize, $valsize) = 
	&Data::Array2ArrayMap::SDB::Writer::TableItem::count_data_offsets(Data::Array2ArrayMap::SDB::Writer::SDB_INT4,$self->{VALTYPE},[$defval]);
	$DEFAULT_VALUE=$self->{DEFAULT_VALUE};
	#print STDERR "valsize=$valsize, DEFAULT_VALUE=$DEFAULT_VALUE\n";
	if ($valsize>0) {
	    $DEFAULT_VALUE = $self->{END_OFFSET};
	    seek ($IDXFILE, $self->{HEADER_OFFSET}+$self->{END_OFFSET}, 0);
	    print $IDXFILE &pack_value($self->{VALTYPE},$self->{DEFAULT_VALUE});
	    $self->{END_OFFSET}+=$valsize;
	}
    }

     my $type=$self->{VALTYPE};

    #			flag	      # bit
    my $flags=$self->{IS_GET_UNIQUE}	<<0
	 | $self->{IS_FIND_UNIQUE}	<<1
	 | 	$HAS_DEFAULT_VALUE	<<2
	 ;

    # creating header
    seek ($IDXFILE, $self->{HEADER_OFFSET}, 0);

    print $IDXFILE NRIF_MAGIC_STRING; # magic value
    print $IDXFILE pack("L",$self->{KEYTYPE});
    print $IDXFILE pack("L",$self->{VALTYPE});
    print $IDXFILE pack("L",$self->{ELNUM});
    print $IDXFILE Data::Array2ArrayMap::SDB::Writer::packoffset($self->{GET_TABLE_OFFSET}); 
    print $IDXFILE Data::Array2ArrayMap::SDB::Writer::packoffset($self->{FIND_TABLE_OFFSET}); 
    print $IDXFILE pack("L",$flags); 
    print $IDXFILE pack("L",$DEFAULT_VALUE); 
}

# private
sub write_data {
    my ($self,$IDXFILE, $orderptr)=@_;
    &__write_data_raw($self,$IDXFILE,$self->{KEYTYPE},$self->{KEYS_OFFSET},&Data::Array2ArrayMap::SDB::Writer::TableItem::convert2key($orderptr));
    &__write_data_raw($self,$IDXFILE,$self->{VALTYPE},$self->{VALUES_OFFSET},&Data::Array2ArrayMap::SDB::Writer::TableItem::convert2val($orderptr));
}

sub __write_data_raw {
    my ($self,$IDXFILE, $TYPE, $offset, $dataptr)=@_;
    return if ! $HAS_EXTERNAL_VALUES{$TYPE};
    print STDERR "wrote values of type $TYPE at $offset\n" if $self->{DEBUG};
    seek ($IDXFILE, $self->{HEADER_OFFSET}+$offset, 0);
    print $IDXFILE map {&pack_value($TYPE,$_)} @$dataptr;
}

# private
sub pack_value {
    my ($TYPE,$value)=@_;
    if ($TYPE==Data::Array2ArrayMap::SDB::Writer::SDB_STRING4) {
	return "$value\0";
    } elsif ($TYPE==Data::Array2ArrayMap::SDB::Writer::SDB_INT8) {
	my $INT4SUP=65536*65536;#=4294967296; # == 2^32
	# we split 8 bytes to pair
	my $low=$value % $INT4SUP;
	my $high=sprintf('%d',$value/$INT4SUP);
	return pack ('L',$low), 
	$value<0 ? 
	    pack ('l',$high-1)
	    : 
	    pack ('L',$high);
    } elsif ($TYPE==Data::Array2ArrayMap::SDB::Writer::SDB_DOUBLE) {
	pack ('d',$value);
    } else {
	# do nothing
    }
}



# private 
sub calc_tables_offsets {
    my ($self)=@_;
    my $type=$self->{VALTYPE};
    my $VALUE_LENGTH=BYTES_IN_OFFSET_T;
    my $KEY_LENGTH=BYTES_IN_OFFSET_T;
#    if ($type == Data::Array2ArrayMap::SDB::Writer::SDB_INT8) {
#	$VALUE_LENGTH=8;
#    }

    my $table_size=($VALUE_LENGTH+$KEY_LENGTH)*$self->{ELNUM};

    if ($self->{HAS_FIND} and $self->{HAS_GET}) {
	$self->{GET_TABLE_OFFSET}=$NRIF_HEADER_OFFSET;
	$self->{FIND_TABLE_OFFSET}=$self->{GET_TABLE_OFFSET}+$table_size;
	$self->{END_OFFSET}=$self->{FIND_TABLE_OFFSET}+$table_size;
    } elsif ($self->{HAS_GET}) {
	$self->{GET_TABLE_OFFSET}=$NRIF_HEADER_OFFSET;
	$self->{END_OFFSET}=$self->{GET_TABLE_OFFSET}+$table_size;
	$self->{FIND_TABLE_OFFSET}=0;
    } elsif ($self->{HAS_FIND}) {
	$self->{FIND_TABLE_OFFSET}=$NRIF_HEADER_OFFSET;
	$self->{END_OFFSET}=$self->{FIND_TABLE_OFFSET}+$table_size;
	$self->{GET_TABLE_OFFSET}=0;
    } else {
	die "broken attribute without get or find!";
    }
    print STDERR "FO=$self->{FIND_TABLE_OFFSET} GO=$self->{GET_TABLE_OFFSET}\n" if $self->{DEBUG};
    $self->{KEYS_OFFSET}=$self->{END_OFFSET};
    $self->{VALUES_OFFSET}=$self->{END_OFFSET};
}

sub write_table {
    my ($self,$IDXFILE, $offset, $keytype, $valtype,$argptr,$valptr)=@_;
    seek ($IDXFILE, $self->{HEADER_OFFSET}+$offset, 0);
    print $IDXFILE $packsub{$keytype}(@$argptr);
    print $IDXFILE $packsub{$valtype}(@$valptr);
}

1;

__END__
# Below is stub documentation for your module. You better edit it!

=head1 NAME

Data::Array2ArrayMap::SDB::Writer - Perl extension for blah blah blah

=head1 SYNOPSIS

  use Data::Array2ArrayMap::Object;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for Data::Array2ArrayMap::Object, created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head2 EXPORT

None by default.


=head1 AUTHOR

A. U. Thor, a.u.thor@a.galaxy.far.far.away

=head1 SEE ALSO

perl(1).

=cut


### Local Variables: 
### mode: perl
### End: 
